"""Tests the pyinilint.py module."""

import configparser
import logging
import unittest

from pyinilint.pyinilint import get_command_line_argument_parser
from pyinilint.pyinilint import get_log_level
from pyinilint.pyinilint import validate_args
from pyinilint.pyinilint import get_config_parser
from pyinilint.pyinilint import readable_paths
from pyinilint.pyinilint import lint
from pyinilint.pyinilint import interpolate
from pyinilint.pyinilint import serialize

GOOD1_PATH = 'pyinilint/tests/good1.ini'
GOOD2_PATH = 'pyinilint/tests/good2.ini'
BAD_PATH = 'pyinilint/tests/bad.ini'


class TestGetCommandLineArgumentParser(unittest.TestCase):
    """ Test the get_command_line_argument_parser() function."""

    def setUp(self):
        """Construct objects used for testing."""
        self.parser = get_command_line_argument_parser()

    def test_basic(self):
        """Test the basic option."""
        self.assertTrue(
            self.parser.parse_args(['-b', GOOD1_PATH]).basic)
        self.assertTrue(
            self.parser.parse_args(['--basic', GOOD1_PATH]).basic)
        self.assertFalse(
            self.parser.parse_args([GOOD1_PATH]).basic)

    def test_debug(self):
        """Test the debug option."""
        self.assertTrue(
            self.parser.parse_args(['-d', GOOD1_PATH]).debug)
        self.assertTrue(
            self.parser.parse_args(['--debug', GOOD1_PATH]).debug)
        self.assertFalse(
            self.parser.parse_args([GOOD1_PATH]).debug)

    def test_encoding(self):
        """Test the encoding option."""
        self.assertEqual(
            self.parser.parse_args(
                ['-e', 'ascii', GOOD1_PATH]
            ).encoding, 'ascii')
        self.assertEqual(
            self.parser.parse_args(
                ['--encoding', 'ascii', GOOD1_PATH]
            ).encoding, 'ascii')
        self.assertIsNone(
            self.parser.parse_args([GOOD1_PATH]).encoding)

    def test_interpolate(self):
        """Test the interpolate option."""
        self.assertTrue(
            self.parser.parse_args(['-i', GOOD1_PATH]).interpolate)
        self.assertTrue(
            self.parser.parse_args(['--interpolate', GOOD1_PATH]).interpolate)
        self.assertFalse(
            self.parser.parse_args([GOOD1_PATH]).interpolate)

    def test_merge(self):
        """Test the merge option."""
        self.assertTrue(
            self.parser.parse_args(['-m', GOOD1_PATH]).merge)
        self.assertTrue(
            self.parser.parse_args(['--merge', GOOD1_PATH]).merge)
        self.assertFalse(
            self.parser.parse_args([GOOD1_PATH]).merge)

    def test_output(self):
        """Test the output option."""
        self.assertTrue(
            self.parser.parse_args(['-o', GOOD1_PATH]).output)
        self.assertTrue(
            self.parser.parse_args(['--output', GOOD1_PATH]).output)
        self.assertFalse(
            self.parser.parse_args([GOOD1_PATH]).output)

    def test_raw(self):
        """Test the raw option."""
        self.assertTrue(
            self.parser.parse_args(['-r', GOOD1_PATH]).raw)
        self.assertTrue(
            self.parser.parse_args(['--raw', GOOD1_PATH]).raw)
        self.assertFalse(
            self.parser.parse_args([GOOD1_PATH]).raw)

    def test_verbose(self):
        """Test the verbose option."""
        self.assertTrue(
            self.parser.parse_args(['-v', GOOD1_PATH]).verbose)
        self.assertTrue(
            self.parser.parse_args(['--verbose', GOOD1_PATH]).verbose)
        self.assertFalse(
            self.parser.parse_args([GOOD1_PATH]).verbose)

    def test_paths(self):
        """Test the paths argument."""
        self.assertEqual(
            self.parser.parse_args(['-v', GOOD1_PATH]).paths,
            [GOOD1_PATH])
        self.assertEqual(
            self.parser.parse_args(['--verbose', GOOD1_PATH]).paths,
            [GOOD1_PATH])
        self.assertEqual(
            self.parser.parse_args([GOOD1_PATH]).paths,
            [GOOD1_PATH])
        self.assertEqual(
            self.parser.parse_args([GOOD1_PATH, 'test/bad.ini']).paths,
            [GOOD1_PATH, 'test/bad.ini'])


class TestGetLogLevel(unittest.TestCase):
    """ Test the get_log_level(args) function."""

    def setUp(self):
        """Construct objects used for testing."""
        self.parser = get_command_line_argument_parser()

    def test_debug(self):
        """Test the logging mode when debug option is used."""
        args = self.parser.parse_args(['--debug', GOOD1_PATH])
        self.assertEqual(get_log_level(args), logging.DEBUG)

    def test_verbose(self):
        """Test the logging mode when verbose option is used."""
        args = self.parser.parse_args(['--verbose', GOOD1_PATH])
        self.assertEqual(get_log_level(args), logging.INFO)

    def test_default(self):
        """Test the logging mode when no debug or verbose option is used."""
        args = self.parser.parse_args([GOOD1_PATH])
        self.assertEqual(get_log_level(args), logging.WARNING)


class TestValidateArgs(unittest.TestCase):
    """ Test the validate_args(args) function."""

    def setUp(self):
        """Construct objects used for testing."""
        self.parser = get_command_line_argument_parser()

    def test_basic_and_raw(self):
        """Test use of basic and raw options causes exception."""
        args = self.parser.parse_args(['--basic', '--raw', GOOD1_PATH])
        with self.assertRaises(ValueError):
            validate_args(args)

    def test_debug_and_verbose(self):
        """Test use of debug and verbose options causes exception."""
        args = self.parser.parse_args(['--debug', '--verbose', GOOD1_PATH])
        with self.assertRaises(ValueError):
            validate_args(args)


class TestReadablePaths(unittest.TestCase):
    """Test the readable_paths(paths) function."""

    def test_good_paths(self):
        """Test readable_paths(paths) when good paths are used."""
        rpaths = readable_paths([GOOD1_PATH, GOOD2_PATH])
        self.assertIn(GOOD1_PATH, rpaths)
        self.assertIn(GOOD2_PATH, rpaths)

    def test_non_existant_paths(self):
        """Test readable_paths(paths) with non-existant path."""
        with self.assertRaises(FileNotFoundError):
            readable_paths(['does/not/exist.ini'])


class TestGetConfigParser(unittest.TestCase):
    """ Test the get_config_parser(args) function."""

    def setUp(self):
        """Construct objects used for testing."""
        self.parser = get_command_line_argument_parser()

    def test_raw(self):
        """Test the config_parser when raw option is used."""
        args = self.parser.parse_args(['--raw', GOOD1_PATH])
        parser = get_config_parser(args)
        parser.read(GOOD1_PATH)
        self.assertEqual(
            parser.get('interpolation-test', 'no-interpolation'),
            'my-value')
        self.assertEqual(
            parser.get('interpolation-test', 'basic-interpolation'),
            '%(my-default)s')
        self.assertEqual(
            parser.get('interpolation-test', 'extended-interpolation'),
            '${my-default}')

    def test_basic(self):
        """Test the config_parser when basic option is used."""
        args = self.parser.parse_args(['--basic', GOOD1_PATH])
        parser = get_config_parser(args)
        parser.read(GOOD1_PATH)
        self.assertEqual(
            parser.get('interpolation-test', 'no-interpolation'),
            'my-value')
        self.assertEqual(
            parser.get('interpolation-test', 'basic-interpolation'),
            '42')
        self.assertEqual(
            parser.get('interpolation-test', 'extended-interpolation'),
            '${my-default}')

    def test_default(self):
        """Test config_parser when default extended interpolation is used."""
        args = self.parser.parse_args([GOOD1_PATH])
        parser = get_config_parser(args)
        parser.read(GOOD1_PATH)
        self.assertEqual(
            parser.get('interpolation-test', 'no-interpolation'),
            'my-value')
        self.assertEqual(
            parser.get('interpolation-test', 'basic-interpolation'),
            '%(my-default)s')
        self.assertEqual(
            parser.get('interpolation-test', 'extended-interpolation'),
            '42')


class TestLint(unittest.TestCase):
    """Test the lint(config, paths, args) function."""

    def setUp(self):
        """Construct objects used for testing."""
        self.parser = get_command_line_argument_parser()

    def test_good_ini_file(self):
        """Test the linting of a good INI file."""
        args = self.parser.parse_args([GOOD1_PATH])
        config = get_config_parser(args)
        rpaths = readable_paths(args.paths)
        new_config = lint(config, rpaths, args)
        self.assertTrue(new_config)

    def test_bad_ini_file(self):
        """Test the linting of a bad INI file."""
        args = self.parser.parse_args([BAD_PATH])
        config = get_config_parser(args)
        rpaths = readable_paths(args.paths)

        with self.assertRaises(configparser.Error):
            lint(config, rpaths, args)


class TestInterpolate(unittest.TestCase):
    """Test the interpolate(config) function."""

    def setUp(self):
        """Construct objects used for testing."""
        self.cfgpsr = configparser.ConfigParser(
            interpolation=configparser.ExtendedInterpolation())

    def test_successful_interpolation(self):
        """Test interpolation of good file."""
        self.cfgpsr.read('pyinilint/tests/extended_interpolation_ok.ini')

        # pylint: disable=W0703
        # Disable catching of too general an exception
        try:
            interpolate(self.cfgpsr)
            # pylint: disable=W1503
            # Disable redundant use of assertTrue(True)
            self.assertTrue(True)
        except Exception as ex:
            self.assertFalse(
                ex,
                "Unexpected exception was raised in "
                "interpolate(config) function. Exception={}"
                .format(ex)
            )

    def test_duplicate_option_error(self):
        """Test interpolation of file with duplicate option error."""
        with self.assertRaises(configparser.DuplicateOptionError):
            self.cfgpsr.read('pyinilint/tests/duplicate_option_error.ini')

    def test_duplicate_section_error(self):
        """Test interpolation of file with duplicate section error."""
        with self.assertRaises(configparser.DuplicateSectionError):
            self.cfgpsr.read('pyinilint/tests/duplicate_section_error.ini')

    def test_interpolation_depth_error(self):
        """Test interpolation of file with interpolation depth error."""
        self.cfgpsr.read('pyinilint/tests/interpolation_depth_error.ini')

        with self.assertRaises(configparser.InterpolationDepthError):
            interpolate(self.cfgpsr)

    def test_interpolation_missing_option_error(self):
        """Test interpolation of file with missing option error."""
        self.cfgpsr.read(
            'pyinilint/tests/interpolation_missing_option_warning.ini')

        with self.assertRaises(configparser.InterpolationMissingOptionError):
            interpolate(self.cfgpsr)

    def test_missing_section_header_error(self):
        """Test interpolation of file with missing section header error."""
        with self.assertRaises(configparser.MissingSectionHeaderError):
            self.cfgpsr.read(
                'pyinilint/tests/missing_section_header_error.ini')

    def test_parsing_error(self):
        """Test interpolation of file with parsing error."""
        with self.assertRaises(configparser.ParsingError):
            self.cfgpsr.read(
                'pyinilint/tests/parsing_error.ini')


class TestSerialize(unittest.TestCase):
    """Test the serialize(config) function."""

    def test_raw_interpolation(self):
        """Test serializing with raw (no) interpolation."""
        config = configparser.ConfigParser(interpolation=None)

        my_section = "my_section"
        my_key = "my_key"
        my_value = "my_value"

        config.add_section(my_section)
        config.set(my_section, my_key, my_value)

        serialized = "[" + my_section + "]\n" + my_key \
            + " = " + my_value + "\n\n"
        self.assertEqual(serialize(config), serialized)


if __name__ == '__main__':
    unittest.main()



"""Lint Python INI-style configuration files."""

import argparse
import configparser
import logging
import os
import sys

__author__ = "Daniel J. R. May"
__copyright__ = "Copyright 2020, Daniel J. R. May"
__credits__ = ["Daniel J. R. May"]
__license__ = "GPLv3"
__version__ = "0.12"
__maintainer__ = "Daniel J. R. May"
__email__ = "daniel.may@danieljrmay.com"
__status__ = "Beta"

EXIT_OK = 0
EXIT_NON_EXISTANT_FILE = 1
EXIT_SYNTAX_ERROR = 2
EXIT_UNREADABLE_FILE = 3
EXIT_DUPLICATE_SECTION = 4
EXIT_DUPLICATE_OPTION = 5
EXIT_INTERPOLATION_MISSING_OPTION = 6
EXIT_INTERPOLATION_DEPTH = 7
EXIT_MISSING_SECTION_HEADER = 8
EXIT_PARSING_ERROR = 9
EXIT_UNKNOWN_ERROR = 10


# pylint: disable=R0912
# pylint: disable=R0915
# Disable too-many-branches and too-many-statements
# It is a reasonable warning as this function is too long, so it
# should be refactored soon.
def main():
    """Entry point"""
    argparser = get_command_line_argument_parser()
    args = argparser.parse_args()

    logging.basicConfig(
        format="[%(levelname)s] %(message)s",
        level=get_log_level(args))
    logging.debug('pyinilint started')
    logging.debug("args=%s", args)

    try:
        validate_args(args)
        rpaths = readable_paths(args.paths)
        config_parser = get_config_parser(args)

        if args.merge:
            lint(config_parser, rpaths, args)
        else:
            for path in rpaths:
                lint(config_parser, path, args)

    except ValueError as val_err:
        logging.error(str(val_err))
        sys.exit(EXIT_SYNTAX_ERROR)
    except FileNotFoundError as fnf_err:
        logging.error(str(fnf_err))
        sys.exit(EXIT_NON_EXISTANT_FILE)
    except PermissionError as perm_err:
        logging.error(str(perm_err))
        sys.exit(EXIT_UNREADABLE_FILE)
    except configparser.DuplicateSectionError as err:
        message = "Section '{}' already exists.".format(err.section)
        error_message = format_error_message(
            "ERROR", "Duplicate section", err.lineno,
            err.source, message)
        print(error_message)
        sys.exit(EXIT_DUPLICATE_SECTION)
    except configparser.DuplicateOptionError as err:
        message = "Option '{}' already exists in section '{}'.".format(
            err.option, err.section)
        error_message = format_error_message(
            "ERROR", "Duplicate option", err.lineno,
            err.source, message)
        print(error_message)
        sys.exit(EXIT_DUPLICATE_OPTION)
    except configparser.InterpolationMissingOptionError as err:
        message = (
            "Option '{}' in section '{}' contains interpolation "
            "key '{}' which does not exist in this fileset.").format(
                err.option, err.section, err.reference)
        error_message = format_error_message(
            "WARNING", "Interpolation missing option", 0,
            rpaths[0], message)
        print(error_message)
        sys.exit(EXIT_INTERPOLATION_MISSING_OPTION)
    except configparser.InterpolationDepthError as err:
        message = ("Recursion limit exceeded in interpolation key '{}' "
                   "in substitution of option '{}' in section '{}'."
                   .format(err.args[2], err.option, err.section))
        error_message = format_error_message(
            "ERROR", "Interpolation depth error", 0,
            rpaths[0], message)
        print(error_message)
        sys.exit(EXIT_INTERPOLATION_MISSING_OPTION)
    except configparser.MissingSectionHeaderError as err:
        message = ("No section header before the line '{}'."
                   .format(err.line.strip()))
        error_message = format_error_message(
            "ERROR", "Missing section header", err.lineno,
            err.source, message)
        print(error_message)
        sys.exit(EXIT_MISSING_SECTION_HEADER)
    except configparser.ParsingError as err:
        error_message = ""
        for line_err in err.errors:
            line_number = line_err[0]
            line = line_err[1]
            message = "Parsing error on line {}.".format(line.strip())
            error_message += "\n" + format_error_message(
                "ERROR", "Parsing error", line_number,
                err.source, message)
        print(error_message)
        sys.exit(EXIT_PARSING_ERROR)
    except configparser.Error as err:
        message = err.message
        error_message = format_error_message(
            "ERROR", "Unknown error", 0,
            rpaths[0], message)
        print(error_message)
        sys.exit(EXIT_UNKNOWN_ERROR)
    finally:
        logging.debug('pyinilint finished')


def get_command_line_argument_parser():
    """Return an ArgumentParser object."""
    description = ('pyinilint (version {}) is a linter and '
                   'inspector for INI format files.'.format(__version__))
    epilog = ('See https://github.com/danieljrmay/pyinilint '
              'for more information.')

    parser = argparse.ArgumentParser(
        prog='pyinilint',
        description=description,
        epilog=epilog
    )
    parser.add_argument(
        '-b', '--basic',
        action='store_true',
        help='use basic interpolation, the default is extended'
    )
    parser.add_argument(
        '-d', '--debug',
        action='store_true',
        help='show debugging messages'
    )
    parser.add_argument(
        '-e', '--encoding',
        help='set the encoding to be used, omit to use the default'
    )
    parser.add_argument(
        '-i', '--interpolate',
        action='store_true',
        help='interpolate the parsed configuration without output'
    )
    parser.add_argument(
        '-m', '--merge',
        action='store_true',
        help='merge files into a single configuration'
    )
    parser.add_argument(
        '-o', '--output',
        action='store_true',
        help='output the parsed configuration to stdout'
    )
    parser.add_argument(
        '-r', '--raw',
        action='store_true',
        help='output raw, do not interpolate'
    )
    parser.add_argument(
        '-s', '--serialize',
        action='store_true',
        help='output the interpolated and serialized configuration to stdout'
    )
    parser.add_argument(
        '-v', '--verbose',
        action='store_true',
        help='show verbose messages'
    )
    parser.add_argument(
        'paths',
        nargs='+',
        help='paths of the file(s) to check'
    )
    return parser


def get_log_level(args):
    """Return the log level requsted by the command line arguments."""
    if args.debug:
        return logging.DEBUG

    if args.verbose:
        return logging.INFO

    return logging.WARNING


def validate_args(args):
    """Validate the parsed command line arguments."""
    if args.basic and args.raw:
        raise ValueError(
            "The -b (--basic) and -r (--raw) options "
            "can not be used together.")

    if args.debug and args.verbose:
        raise ValueError(
            "The -d (--debug) and -v (--verbose) options "
            "can not be used together.")


def readable_paths(paths):
    """Get a list of readable paths."""
    rpaths = []

    for path in paths:
        if not os.path.isfile(path):
            raise FileNotFoundError("File not found at path {}".format(path))

        if not os.access(path, os.R_OK):
            raise PermissionError("Unreadable file at path {}".format(path))

        logging.debug("Readable file at path %s", path)
        rpaths.append(path)

    return rpaths


def get_config_parser(args):
    """Return the config parser."""
    if args.raw:
        return configparser.ConfigParser(interpolation=None)

    if args.basic:
        return configparser.ConfigParser(
            interpolation=configparser.BasicInterpolation())

    return configparser.ConfigParser(
        interpolation=configparser.ExtendedInterpolation())


def lint(config, paths, args):
    """Lint the INI-style files at the specified paths."""
    logging.debug("About to read paths %s", paths)
    linted_paths = config.read(
        paths,
        args.encoding)

    if args.verbose:
        print("\n".join(linted_paths))

    if args.interpolate:
        logging.debug("Interpolating config:")
        interpolate(config)

    if args.output:
        logging.debug("Printing uninterpolated config:")
        config.write(sys.stdout)

    if args.serialize:
        logging.debug("Serializing interpolated config:")
        print(serialize(config))

    return config


def interpolate(config):
    """Interpolate the parsed config, but do not output anything."""
    for section in config.sections():
        options = config.options(section)
        for option in options:
            config.get(section, option)


def serialize(config):
    """Return the interpolated and serialized config."""
    output = ""

    for section in config.sections():
        output += "[" + section + "]\n"

        options = config.options(section)
        for option in options:
            output += option + " = " + config.get(section, option) + "\n"

        output += "\n"

    return output


def format_error_message(level, identity, line_number, filename, message):
    """Return an error message suitable for flycheck-ini-pyinilint."""
    return "[{}] {} at line {} of {}: {}".format(
        level, identity, line_number, filename, message)


if __name__ == "__main__":
    main()



"""setuptools configuration for the pyinilint package.

See: https://packaging.python.org/tutorials/packaging-projects
"""

import setuptools

with open("README.md", "r") as fh:
    LONG_DESCRIPTION = fh.read()

setuptools.setup(
    name="pyinilint",
    version="0.16",
    author="Daniel J. R. May",
    author_email="daniel.may@danieljrmay.com",
    description="A linter and inspector for INI format files",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/danieljrmay/pyinilint",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
    ],
    entry_points={
        "console_scripts": ["pyinilint=pyinilint.pyinilint:main"],
    },
    test_suite="pyinilint.tests",
    include_package_data=False,
)
